@@ -1,7 +1,7 @@
 /// @file htslib/faidx.h
 /// FASTA random access.
 /*
-   Copyright (C) 2008, 2009, 2013, 2014, 2016, 2017-2020 Genome Research Ltd.
+   Copyright (C) 2008, 2009, 2013, 2014, 2016, 2017-2020, 2022-2023 Genome Research Ltd.
 
    Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -70,6 +70,9 @@ struct faidx_t;
 /// Opaque structure representing FASTA index
 typedef struct faidx_t faidx_t;
 
+/// Opaque structure; sole item needed from htslib/thread_pool.h
+struct hts_tpool;
+
 /// File format to be dealing with.
 enum fai_format_options {
     FAI_NONE,
@@ -188,6 +191,15 @@ char *fai_fetch(const faidx_t *fai, const char *reg, int *len);
 HTSLIB_EXPORT
 char *fai_fetch64(const faidx_t *fai, const char *reg, hts_pos_t *len);
 
+/// Query the line-wrap length for a chromosome specified as part of a region
+/** @param  fai  Pointer to the faidx_t struct
+    @param  reg  Region in the format "chr2:20,000-30,000"
+    @return      The line length (excluding newline),
+                 negative on error.
+*/
+HTSLIB_EXPORT
+hts_pos_t fai_line_length(const faidx_t *fai, const char *reg);
+
 /// Fetch the quality string for a region for FASTQ files
 /** @param  fai  Pointer to the faidx_t struct
     @param  reg  Region in the format "chr2:20,000-30,000"
@@ -283,7 +295,22 @@ int faidx_nseq(const faidx_t *fai);
 HTSLIB_EXPORT
 const char *faidx_iseq(const faidx_t *fai, int i);
 
-/// Return sequence length, -1 if not present
+/// Return sequence length
+/** @param  fai  Pointer to the faidx_t struct
+    @param  seq  Name of the sequence
+    @return Sequence length, or -1 if not present
+*/
+HTSLIB_EXPORT
+hts_pos_t faidx_seq_len64(const faidx_t *fai, const char *seq);
+
+/// Return sequence length
+/** @param  fai  Pointer to the faidx_t struct
+    @param  seq  Name of the sequence
+    @return Sequence length, or -1 if not present
+
+    @deprecated This funtion cannot handle very long sequences.
+                Use faidx_seq_len64() instead.
+*/
 HTSLIB_EXPORT
 int faidx_seq_len(const faidx_t *fai, const char *seq);
 
@@ -305,6 +332,27 @@ const char *fai_parse_region(const faidx_t *fai, const char *s,
                              int *tid, hts_pos_t *beg, hts_pos_t *end,
                              int flags);
 
+/// Adjust region to the actual sequence length
+/** @param  fai   Pointer to the faidx_t struct
+    @param  tid   Sequence index, as returned by fai_parse_region()
+    @param  beg[in,out]   The start of the region (0 based)
+    @param  end[in,out]   One past end of the region (0 based)
+    @return 1, 2, or 3 if @p beg, @p end, or both are adjusted,
+            0 if @p beg and @p end are unchanged
+            -1 on error
+
+    Looks up the length of @p tid, and then adjusts the values of @p beg
+    and @p end if they fall outside the boundaries of the sequence.
+
+    If @p beg > @p end, it will be set to @p end.
+
+    The return value indicates which, if any, of the inputs have been
+    adjusted.  -1 will be returned if @p tid is not a valid sequence index.
+*/
+HTSLIB_EXPORT
+int fai_adjust_region(const faidx_t *fai, int tid,
+                      hts_pos_t *beg, hts_pos_t *end);
+
 /// Sets the cache size of the underlying BGZF compressed file
 /** @param  fai         Pointer to the faidx_t struct
  *  @param  cache_size  Selected cache size in bytes
@@ -312,6 +360,15 @@ const char *fai_parse_region(const faidx_t *fai, const char *s,
 HTSLIB_EXPORT
 void fai_set_cache_size(faidx_t *fai, int cache_size);
 
+/// Adds a thread pool to the underlying BGZF layer.
+/** @param fai         FAI file handler
+ *  @param pool        The thread pool (see hts_create_threads)
+ *  @param qsize       The size of the job queue.  If 0 this is twice the
+ *                     number of threads in the pool.
+ */
+HTSLIB_EXPORT
+int fai_thread_pool(faidx_t *fai, struct hts_tpool *pool, int qsize);
+
 /// Determines the path to the reference index file
 /** @param  fa    String with the path to the reference file
  *  @return       String with the path to the reference index file, or NULL on failure
@@ -347,7 +404,7 @@ HTSLIB_EXPORT
 int faidx_fetch_seq2(const faidx_t *fai, const char *c_name, int p_beg_i, int p_end_i, char *out);
 
 /// Alternative to faidx_fetch_seq() -- originally by Daniel Jones (author of
-/// the seqbias package), moved to Rhtslib and ported to htslib-1.15.1 by H.P.
+/// the seqbias package), moved to Rhtslib and ported to htslib-1.18 by H.P.
 /** @param  fai  Pointer to the faidx_t struct
     @param  c_name Region name
     @param  p_beg_i  Beginning position number (zero-based)
