@@ -57,7 +57,7 @@ int kputd(double d, kstring_t *s) {
 		if (ks_resize(s, s->l + 50) < 0)
 			return EOF;
 		// We let stdio handle the exponent cases
-		int s2 = sprintf(s->s + s->l, "%g", d);
+		int s2 = snprintf(s->s + s->l, s->m - s->l, "%g", d);
 		len += s2;
 		s->l += s2;
 		return len;
@@ -166,24 +166,33 @@ int kvsprintf(kstring_t *s, const char *fmt, va_list ap)
 		return l;
 	}
 
+	if (!s->s) {
+		const size_t sz = 64;
+		s->s = malloc(sz);
+		if (!s->s)
+			return -1;
+		s->m = sz;
+		s->l = 0;
+	}
+
 	l = vsnprintf(s->s + s->l, s->m - s->l, fmt, args); // This line does not work with glibc 2.0. See `man snprintf'.
-	va_end(args);
-	int increase = 64;                                          /* H.P. */
-	while (l == -1 || l + 1 > s->m - s->l) {                    /* H.P. */
-		if (l != -1) {                                      /* H.P. */
-			if (ks_resize(s, s->l + l + 2) < 0)         /* H.P. */
-				return -1;                          /* H.P. */
-		} else {                                            /* H.P. */
-			increase *= 2;                              /* H.P. */
-			if (ks_resize(s, s->m + increase) < 0)      /* H.P. */
-				return -1;                          /* H.P. */
-		}                                                   /* H.P. */
-		va_copy(args, ap);                                  /* H.P. */
-		l = vsnprintf(s->s + s->l, s->m - s->l, fmt, args); /* H.P. */
-		va_end(args);                                       /* H.P. */
-		if (l != -1)                                        /* H.P. */
-			break;                                      /* H.P. */
-	}                                                           /* H.P. */
+        va_end(args);
+        int increase = 64;                                          /* H.P. */
+        while (l == -1 || l + 1 > s->m - s->l) {                    /* H.P. */
+                if (l != -1) {                                      /* H.P. */
+                        if (ks_resize(s, s->l + l + 2) < 0)         /* H.P. */
+                                return -1;                          /* H.P. */
+                } else {                                            /* H.P. */
+                        increase *= 2;                              /* H.P. */
+                        if (ks_resize(s, s->m + increase) < 0)      /* H.P. */
+                                return -1;                          /* H.P. */
+                }                                                   /* H.P. */
+                va_copy(args, ap);                                  /* H.P. */
+                l = vsnprintf(s->s + s->l, s->m - s->l, fmt, args); /* H.P. */
+                va_end(args);                                       /* H.P. */
+                if (l != -1)                                        /* H.P. */
+                        break;                                      /* H.P. */
+        }                                                           /* H.P. */
 	s->l += l;
 	return l;
 }
