@@ -1,6 +1,6 @@
 #!/usr/bin/env perl
 #
-#    Copyright (C) 2012-2021 Genome Research Ltd.
+#    Copyright (C) 2012-2023 Genome Research Ltd.
 #
 #    Author: Petr Danecek <pd3@sanger.ac.uk>
 #
@@ -32,6 +32,7 @@ use File::Temp qw/ tempfile tempdir /;
 use IO::Handle;
 
 my $opts = parse_params();
+srand($$opts{seed});
 
 test_bgzip($opts, 0);
 test_bgzip($opts, 4);
@@ -53,12 +54,15 @@ test_bcf2vcf($opts);
 test_vcf_sweep($opts,out=>'test-vcf-sweep.out');
 test_vcf_various($opts);
 test_bcf_sr_sort($opts);
+test_bcf_sr_no_index($opts);
+test_bcf_sr_range($opts);
 test_command($opts,cmd=>'test-bcf-translate -',out=>'test-bcf-translate.out');
 test_convert_padded_header($opts);
 test_rebgzip($opts);
 test_logging($opts);
 test_plugin_loading($opts);
 test_realn($opts);
+test_bcf_set_variant_type($opts);
 
 print "\nNumber of tests:\n";
 printf "    total   .. %d\n", $$opts{nok}+$$opts{nfailed};
@@ -79,6 +83,7 @@ sub error
         "Usage: test.pl [OPTIONS]\n",
         "Options:\n",
         "   -r, --redo-outputs              Recreate expected output files.\n",
+        "   -s, --random-seed <int>         Initialise rand() with a different seed.\n",
         "   -t, --temp-dir <path>           When given, temporary files will not be removed.\n",
         "   -f, --fail-fast                 Fail-fast mode: exit as soon as a test fails.\n",
         "   -h, -?, --help                  This help message.\n",
@@ -104,12 +109,13 @@ sub safe_tempdir
 
 sub parse_params
 {
-    my $opts = { keep_files=>0, nok=>0, nfailed=>0 };
+    my $opts = { keep_files=>0, nok=>0, nfailed=>0, seed=>42 };
     my $help;
     Getopt::Long::Configure('bundling');
     my $ret = GetOptions (
             't|temp-dir:s' => \$$opts{keep_files},
             'r|redo-outputs' => \$$opts{redo_outputs},
+            's|random-seed=i' => \$$opts{seed},
             'f|fail-fast' => \$$opts{fail_fast},
             'h|?|help' => \$help
             );
@@ -394,6 +400,30 @@ sub test_bgzip {
     }
     passed($opts,$test);
 
+    # Round-trip test of text in binary mode
+    $test = sprintf('%s %2s threads', 'bgzip text mode round-trip',
+                       $threads ? $threads : 'no');
+    print "$test: ";
+    $c = "$$opts{bin}/bgzip $at --binary -i -I '$index' < '$data' > '$compressed'";
+    ($ret, $out) = _cmd($c);
+    if ($ret) {
+        failed($opts, $test, "non-zero exit from $c");
+        return;
+    }
+    $c = "$$opts{bin}/bgzip $at -d < '$compressed' > '$uncompressed'";
+    ($ret, $out) = _cmd($c);
+    if ($ret) {
+        failed($opts, $test, "non-zero exit from $c");
+        return;
+    }
+    $c = "cmp '$data' '$uncompressed'";
+    ($ret, $out) = _cmd($c);
+    if ($ret) {
+        failed($opts, $test, $out ? $out : "'$data' '$uncompressed' differ");
+        return;
+    }
+    passed($opts,$test);
+
     # Extract from an offset
     $test = sprintf('%s %2s threads', 'bgzip -b',
                     $threads ? $threads : 'no');
@@ -614,7 +644,7 @@ sub test_view
 
             ## Experimental CRAM 4.0 support.
             # SAM -> CRAM40 -> SAM
-            my @p = $sam eq "ce#large_seq.sam" || $sam eq "xx#large_aux.sam"
+            @p = $sam eq "ce#large_seq.sam" || $sam eq "xx#large_aux.sam"
                 ? (qw/fast normal small archive/)
                 : (qw/archive/);
             foreach my $profile (@p) {
@@ -643,6 +673,21 @@ sub test_view
         }
     }
 
+    # embed_ref=2 mode
+    print "test_view testing embed_ref=2:\n";
+    $test_view_failures = 0;
+    my $ersam = "ce#1000.sam";
+    my $ercram = "ce#1000_er.tmp.cram";
+    my $ersam2 = "${ercram}.sam";
+    testv $opts, "./test_view $tv_args -C -p $ercram $ersam";
+    testv $opts, "./test_view $tv_args -p $ersam2 $ercram";
+    testv $opts, "./compare_sam.pl $ersam $ersam2";
+    if ($test_view_failures == 0) {
+        passed($opts, "embed_ref=2 tests");
+    } else {
+        failed($opts, "embed_ref=2 tests", "$test_view_failures subtests failed");
+    }
+
     # BAM and CRAM range queries on prebuilt BAM and CRAM
     # The cram file has @SQ UR: set to point to an invalid location to
     # force the reference to be reloaded from the one given on the
@@ -904,6 +949,11 @@ sub test_vcf_various
         cmd => "$$opts{bin}/htsfile -c $$opts{path}/formatmissing.vcf");
     test_cmd($opts, %args, out => "vcf_meta_meta.vcf",
         cmd => "$$opts{bin}/htsfile -c $$opts{path}/vcf_meta_meta.vcf");
+
+    # VCF file with contig IDX=1, simulating an edited BCF file
+    # See htslib issue 1534
+    test_cmd($opts, %args, out => "modhdr.expected.vcf",
+        cmd => "$$opts{path}/test_view $$opts{path}/modhdr.vcf.gz chr22:1-2");
 }
 
 sub write_multiblock_bgzf {
@@ -974,7 +1024,7 @@ sub test_bcf_sr_sort
     my ($opts, %args) = @_;
     for (my $i=0; $i<10; $i++)
     {
-        my $seed = int(rand(time));
+        my $seed = int(rand(100000000));
         my $test = 'test-bcf-sr';
         my $cmd  = "$$opts{path}/test-bcf-sr.pl -t $$opts{tmp} -s $seed";
         print "$test:\n";
@@ -985,6 +1035,113 @@ sub test_bcf_sr_sort
     }
 }
 
+sub test_bcf_sr_no_index {
+    my ($opts) = @_;
+
+    my $test = "test_bcf_sr_no_index";
+
+    my $vcfdir = "$$opts{path}/bcf-sr";
+
+    # Positive test
+    test_cmd($opts, out => "bcf-sr/merge.noidx.abc.expected.out",
+             cmd => "$$opts{path}/test-bcf-sr --no-index -p all --args $vcfdir/merge.noidx.a.vcf $vcfdir/merge.noidx.b.vcf $vcfdir/merge.noidx.c.vcf 2> $$opts{tmp}/no_index_1.err");
+
+    # Check bad input detection
+
+    my @bad_file_tests = (["out-of-order header",
+                           ["merge.noidx.a.vcf", "merge.noidx.hdr_order.vcf"]],
+                          ["out-of-order records",
+                           ["merge.noidx.a.vcf", "merge.noidx.rec_order.vcf"]],
+                          ["out-of-order records",
+                           ["merge.noidx.rec_order.vcf", "merge.noidx.a.vcf"]]);
+    my $count = 2;
+    foreach my $test_params (@bad_file_tests) {
+        my ($badness, $inputs) = @$test_params;
+        my @ins = map { "$vcfdir/$_" } @$inputs;
+
+        my $cmd = "$$opts{path}/test-bcf-sr --no-index -p all --args @ins > $$opts{tmp}/no_index_$count.out 2> $$opts{tmp}/no_index_$count.err";
+        print "$test:\n\t$cmd (expected fail)\n";
+        my ($ret) = _cmd($cmd);
+        if ($ret == 0) {
+            failed($opts, $test, "Failed to detect $badness: $cmd\n");
+        } else {
+            passed($opts, $test);
+        }
+        $count++;
+    }
+}
+
+sub test_bcf_sr_range {
+    my ($opts) = @_;
+
+    my $test = "test_bcf_sr_range";
+
+    my $vcfdir = "$$opts{path}/bcf-sr";
+
+    my @tests = (['r', '1', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['r', '1:1-2', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['r', '1:1,1:2', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['r', '1:1-1', 'weird-chr-names.vcf', 'weird-chr-names.2.out'],
+                 ['r', '{1:1}', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['r', '{1:1}:1-2', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['r', '{1:1}:1,{1:1}:2', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['r', '{1:1}:1-1', 'weird-chr-names.vcf', 'weird-chr-names.4.out'],
+                 ['r', '{1:1-1}', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['r', '{1:1-1}:1-2', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['r', '{1:1-1}:1,{1:1-1}:2', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['r', '{1:1-1}:1-1', 'weird-chr-names.vcf', 'weird-chr-names.6.out'],
+                 ['r', '{1:1-1}-2', 'weird-chr-names.vcf', undef], # Expected failure
+                 ['t', '1', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['t', '1:1-2', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['t', '1:1,1:2', 'weird-chr-names.vcf', 'weird-chr-names.1.out'],
+                 ['t', '1:1-1', 'weird-chr-names.vcf', 'weird-chr-names.2.out'],
+                 ['t', '{1:1}', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['t', '{1:1}:1-2', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['t', '{1:1}:1,{1:1}:2', 'weird-chr-names.vcf', 'weird-chr-names.3.out'],
+                 ['t', '{1:1}:1-1', 'weird-chr-names.vcf', 'weird-chr-names.4.out'],
+                 ['t', '{1:1-1}', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['t', '{1:1-1}:1-2', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['t', '{1:1-1}:1,{1:1-1}:2', 'weird-chr-names.vcf', 'weird-chr-names.5.out'],
+                 ['t', '{1:1-1}:1-1', 'weird-chr-names.vcf', 'weird-chr-names.6.out'],
+                 ['t', '{1:1-1}-2', 'weird-chr-names.vcf', undef] # Expected failure
+        );
+
+    my $count = 0;
+    my %converted;
+    foreach my $tst (@tests) {
+        my ($option, $range, $in, $exp_out) = @$tst;
+        $count++;
+        if (!$converted{$in}) {
+            my $cmd = "$$opts{path}/test_view -b -p $$opts{tmp}/$in.bcf -x $$opts{tmp}/$in.bcf.csi $vcfdir/$in";
+            print "$test:\n\t$cmd\n";
+            my ($ret) = _cmd($cmd);
+            if ($ret) {
+                failed($opts, $test);
+                $converted{$in} = 'fail';
+                next;
+            } else {
+                passed($opts, $test);
+                $converted{$in} = "$$opts{tmp}/$in.bcf";
+            }
+        }
+        next if ($converted{$in} eq 'fail');
+        my $cmd = "$$opts{path}/test-bcf-sr -O vcf -o $$opts{tmp}/range_test_$count.out.vcf -$option '$range' --args $converted{$in}";
+        if ($exp_out) {
+            test_compare($opts, $cmd, "$vcfdir/$exp_out",
+                         "$$opts{tmp}/range_test_$count.out.vcf",
+                         fix_newlines => 1);
+        } else {
+            print "$test:\n\t$cmd (expected fail)\n";
+            my ($ret) = _cmd($cmd);
+            if ($ret) {
+                passed($opts, $test);
+            } else {
+                failed($opts, $test);
+            }
+        }
+    }
+}
+
 sub test_command
 {
     my ($opts, %args) = @_;
@@ -1051,4 +1208,21 @@ sub test_realn {
 
     # Revert quality values (using data in ZQ tags)
     test_cmd($opts, cmd => "$test_realn -f $$opts{path}/realn02.fa -i $$opts{path}/realn02_exp-a.sam -o -", out => "realn02_exp.sam");
+
+    # Make sure multiple matches are treated the same way as a single match of the same length.
+    test_cmd($opts, cmd => "$test_realn -f $$opts{path}/realn03.fa -e -i $$opts{path}/realn03.sam -o -", out => "realn03_exp.sam");
+}
+
+sub test_bcf_set_variant_type
+{
+    my ($opts) = @_;
+    my $test = 'test-bcf_set_variant_type';
+    my $cmd  = "$$opts{path}/test-bcf_set_variant_type";
+    print "$test:\n";
+    print "\t$cmd\n";
+    my ($ret,$out) = _cmd($cmd);
+    if ( $ret ) {
+        print $out;
+        failed($opts,$test);
+    } else { passed($opts,$test); }
 }
